theory Proof_System_1
  imports Main
begin

typedecl State
axiomatization 
trans :: "State \<Rightarrow> State \<Rightarrow> bool" (infixl "\<mapsto>" 500)

type_synonym  SymState = "State \<Rightarrow> bool"

definition final :: "SymState" ("\<bullet>")
  where "\<bullet> \<equiv> \<lambda>s. \<forall> s'. \<not> s \<mapsto> s'" 

definition top :: "SymState"  ("\<top>") where "\<top> \<equiv> \<lambda>_.True"
definition bot :: "SymState"  ("\<bottom>") where "\<bottom> \<equiv> \<lambda>_.False" 
definition cnj :: "SymState \<Rightarrow> SymState \<Rightarrow> SymState" (infixl "\<sqinter>" 200)
  where "p \<sqinter> q \<equiv> (\<lambda> s. p s \<and> q s)"
definition dsj :: "SymState \<Rightarrow> SymState \<Rightarrow> SymState" (infixl "\<squnion>" 220)
  where "p \<squnion> q \<equiv> (\<lambda> s. p s \<or> q s)"
definition imp :: "SymState \<Rightarrow> SymState \<Rightarrow> bool" (infix "\<sqsubseteq>" 180)
  where "p \<sqsubseteq> q \<equiv> (\<forall> s. p s \<longrightarrow> q s)"

definition symTrans :: "SymState \<Rightarrow> SymState" ("\<partial>") where
"\<partial>p \<equiv> \<lambda> s.\<exists> s'. p s' \<and> s' \<mapsto> s" 

codatatype seq = is_one: one (\<alpha>: State) ("\<lparr>_\<rparr>")
  |is_add: add (\<alpha> :State) (\<omega>:seq) ("_\<Colon>_")
               
inductive isFinite :: "seq \<Rightarrow> bool" ("\<propto>")  where
finite_one : " \<propto> \<lparr>s\<rparr>" |
finite_add : "\<propto> \<tau> \<Longrightarrow> \<propto> (s \<Colon> \<tau>)"

definition some ("\<lfloor>_\<rfloor>") where "\<lfloor>x\<rfloor> \<equiv> Some x"
definition none ("\<lfloor>\<rfloor>") where "\<lfloor>\<rfloor> \<equiv> None"

coinductive isPath ("\<pi>") where 
path_one : "\<bullet> s \<Longrightarrow> \<pi> \<lparr>s\<rparr>" |
path_add : "s \<mapsto> \<alpha> \<tau> \<Longrightarrow> \<pi> \<tau> \<Longrightarrow> \<pi> (s \<Colon> \<tau>)"

coinductive reach :: "seq \<Rightarrow> SymState \<Rightarrow> bool" (infix "\<leadsto>" 300)where
reach_now_one : "r s \<Longrightarrow>  \<lparr>s\<rparr> \<leadsto> r" |
reach_now_add : "r s \<Longrightarrow> (s \<Colon> _) \<leadsto> r" |
reach_later : "\<tau> \<leadsto> r \<Longrightarrow>  ( _\<Colon> \<tau>) \<leadsto> r"


datatype reachability_formula =
 arrow (L:SymState) (R:SymState) (infix "\<Rightarrow>\<diamondsuit>" 150)

definition valid :: "reachability_formula \<Rightarrow> bool" ("\<Turnstile>") where
"\<Turnstile> f \<equiv> \<forall> \<tau>. (L f) (\<alpha> \<tau>) \<longrightarrow> \<pi> \<tau> \<longrightarrow>\<propto> \<tau> \<longrightarrow> \<tau> \<leadsto>(R f)"

coinductive prove :: "reachability_formula \<Rightarrow> bool"  ("\<turnstile>" 1000)  
  where
  stp :   "l \<sqsubseteq> (l' \<squnion> r) \<Longrightarrow>
           l' \<sqinter> \<bullet> \<sqsubseteq> \<bottom> \<Longrightarrow>
           \<turnstile> (\<partial> l' \<Rightarrow>\<diamondsuit> r) \<Longrightarrow>
           \<turnstile>(l \<Rightarrow>\<diamondsuit> r)"

lemma sound_aux : "\<turnstile> (l \<Rightarrow>\<diamondsuit>r) \<Longrightarrow>  l (\<alpha> \<tau>) \<Longrightarrow> \<pi> \<tau> \<Longrightarrow> \<propto> \<tau> \<Longrightarrow> \<tau> \<leadsto> r"
proof (coinduction arbitrary: l \<tau>)
case reach thus ?case 
  proof (cases \<tau>)
  case (one st) thus ?thesis 
  proof -    
    have "r st" (* takes a while*)
      by (metis Proof_System_1.seq.distinct(1) bot_def cnj_def dsj_def imp_def isPath.simps one 
      prove.simps reach(1) reach(2) reach(3) reachability_formula.sel(1) reachability_formula.sel(2)
      seq.sel(1))     
    thus ?thesis by (simp add: one) 
  qed
  next case (add st \<tau>') thus ?thesis 
  proof (cases "r st")
    case True thus ?thesis using add by blast
    next case False thus ?thesis 
      proof - 
      obtain l where "\<turnstile> (l \<Rightarrow>\<diamondsuit> r) \<and> l (\<alpha> \<tau>') \<and> \<pi> \<tau>' \<and>  \<propto> \<tau>'"
       by (smt False Proof_System_1.seq.distinct(1) add dsj_def imp_def isFinite.cases isPath.simps
               prove.simps reach reachability_formula.sel seq.sel symTrans_def) 
      thus ?thesis using add by auto
     qed
   qed
  qed
qed

theorem soundness: "\<turnstile> (l  \<Rightarrow>\<diamondsuit> r) \<Longrightarrow> \<Turnstile> ( l \<Rightarrow>\<diamondsuit> r)"
  by (simp add: sound_aux valid_def)


lemma custom_coind : 
"f \<in> C  \<Longrightarrow>
  (\<And> l r. (l \<Rightarrow>\<diamondsuit> r) \<in> C \<Longrightarrow>  
    \<exists>l'. l \<sqsubseteq> (l' \<squnion> r) \<and>  (l' \<sqinter> \<bullet>) \<sqsubseteq> \<bottom> \<and> 
      ((\<partial> l' \<Rightarrow>\<diamondsuit> r) \<in> C  \<or> \<turnstile> (\<partial> l' \<Rightarrow>\<diamondsuit> r))) \<Longrightarrow>
\<turnstile> f"
 using prove.coinduct [of "\<lambda>f. f \<in> C" f]
  by (metis reachability_formula.exhaust) 


lemma strategy :
 "l \<sqsubseteq> (q \<squnion> r) \<Longrightarrow> 
  (q \<sqinter> \<bullet>) \<sqsubseteq> \<bottom> \<Longrightarrow> 
  \<partial> q \<sqsubseteq> (q \<squnion> r) \<Longrightarrow> 
  \<turnstile> (l \<Rightarrow>\<diamondsuit> r)"
proof (coinduction arbitrary:l)
  case prove
  then show ?case
    by blast
qed
 

definition coReach :: "SymState \<Rightarrow> SymState" where
"coReach r \<equiv> \<lambda> s.\<not>r s \<and> (\<forall> \<tau>. s = (\<alpha> \<tau>) \<longrightarrow> \<pi> \<tau> \<longrightarrow>  \<propto> \<tau> \<longrightarrow> \<tau> \<leadsto> r)"

lemma init : "\<Turnstile> (l \<Rightarrow>\<diamondsuit> r) \<Longrightarrow> l \<sqsubseteq> (coReach r \<squnion> r)"
  using coReach_def dsj_def imp_def valid_def by auto


lemma nonBlock : "(coReach r  \<sqinter> \<bullet>) \<sqsubseteq> \<bottom>"
proof (cases "\<exists>s. (coReach r) s \<and> \<bullet> s")
  case True
  thus ?thesis 
  proof -
    have False
      by (metis seq.distinct(1) True coReach_def finite_one path_one reach.simps seq.sel(1))
    thus ?thesis by simp
  qed
next
  case False
  thus ?thesis
    using cnj_def imp_def by auto 
qed


lemma quasiInv : "\<partial> (coReach r) \<sqsubseteq> (coReach r \<squnion> r)"
proof (cases "\<exists> st. (\<partial>(coReach r)) st \<and> \<not> (coReach r) st \<and> \<not> r st ")
  case True
  then show ?thesis
  proof -
    obtain st where "\<partial> (coReach r) st" and  "\<not> (coReach r) st"  and  "\<not> r st" using True 
      by blast
    then obtain \<tau> where "st = (\<alpha> \<tau>)" and "\<pi> \<tau>" and " \<propto> \<tau>" and "\<not> \<tau> \<leadsto> r"
      using coReach_def by auto
    obtain st' where "(coReach r) st'" and "st' \<mapsto> st"
      using \<open>\<partial> (coReach r) st\<close> symTrans_def by auto
    have "\<not> r st'" and "\<pi> (add st' \<tau>)" and "\<propto> (add st' \<tau>)"
      proof -
        show "\<not> r st'"
        using \<open>coReach r st'\<close> coReach_def by force
      next show "isPath (add st' \<tau>)"
          using \<open>trans st' st\<close> \<open>\<pi> \<tau>\<close> \<open>st = \<alpha> \<tau>\<close> path_add by blast
      next show "isFinite (add st' \<tau>)"
          by (simp add: \<open>\<propto> \<tau>\<close> finite_add)
    qed
    have "(add st' \<tau>)\<leadsto> r"
      using \<open>\<pi> (st'\<Colon>\<tau>)\<close> \<open>\<propto> (st'\<Colon>\<tau>)\<close> \<open>coReach r st'\<close> coReach_def by auto
    have False
      using \<open>\<not> \<tau> \<leadsto> r\<close> \<open>\<not> r st'\<close> \<open>(st'\<Colon>\<tau>) \<leadsto> r\<close> reach.simps by blast
    thus ?thesis by auto
  qed
next
case False
  then show ?thesis
    using dsj_def imp_def by auto 
qed


theorem completeness : "\<Turnstile> f \<Longrightarrow> \<turnstile> f "
  by (metis init nonBlock quasiInv reachability_formula.exhaust strategy) 

end
