theory Proof_System_2
  imports Main 
begin

typedecl State
axiomatization 
trans :: "State \<Rightarrow> State \<Rightarrow> bool" (infixl "\<mapsto>" 500)

type_synonym  SymState = "State \<Rightarrow> bool"

definition final :: "SymState" ("\<bullet>")
  where "\<bullet> \<equiv> \<lambda>s. \<forall> s'. \<not> s \<mapsto> s'" 

definition top :: "SymState"  ("\<top>") where "\<top> \<equiv> \<lambda>_.True"
definition bot :: "SymState"  ("\<bottom>") where "\<bottom> \<equiv> \<lambda>_.False" 
definition cnj :: "SymState \<Rightarrow> SymState \<Rightarrow> SymState" (infixl "\<sqinter>" 200)
  where "p \<sqinter> q \<equiv> (\<lambda> s. p s \<and> q s)"
definition dsj :: "SymState \<Rightarrow> SymState \<Rightarrow> SymState" (infixl "\<squnion>" 220)
  where "p \<squnion> q \<equiv> (\<lambda> s. p s \<or> q s)"
definition imp :: "SymState \<Rightarrow> SymState \<Rightarrow> bool" (infix "\<sqsubseteq>" 180)
  where "p \<sqsubseteq> q \<equiv> (\<forall> s. p s \<longrightarrow> q s)"

definition symTrans :: "SymState \<Rightarrow> SymState" ("\<partial>") where
"\<partial>p \<equiv> \<lambda> s.\<exists> s'. p s' \<and> s' \<mapsto> s" 

codatatype seq = is_one: one (\<alpha>: State) ("\<lparr>_\<rparr>")
  |is_add: add (\<alpha> :State) (\<omega>:seq) ("_\<Colon>_")
               

inductive isFinite :: "seq \<Rightarrow> bool" ("\<propto>")  where
finite_one : " \<propto> \<lparr>s\<rparr>" |
finite_add : "\<propto> \<tau> \<Longrightarrow> \<propto> (s \<Colon> \<tau>)"

definition some ("\<lfloor>_\<rfloor>") where "\<lfloor>x\<rfloor> \<equiv> Some x"
definition none ("\<lfloor>\<rfloor>") where "\<lfloor>\<rfloor> \<equiv> None"

fun nth :: "seq \<Rightarrow> nat \<Rightarrow> State option"  ("_\<langle>_\<rangle>"  [800,0] 900) where
"\<tau>\<langle>0\<rangle> = \<lfloor>(\<alpha> \<tau>)\<rfloor>" |
"\<lparr> _\<rparr>\<langle>Suc n\<rangle> = \<lfloor>\<rfloor>" |
"(_\<Colon> \<tau>)\<langle>Suc n\<rangle> = \<tau>\<langle>n\<rangle>"

fun suffix :: "nat \<Rightarrow> seq \<Rightarrow> seq option" ("\<langle>_\<rangle>_"  [0,1000] 999) where
"\<langle>0\<rangle> \<tau> = \<lfloor>\<tau>\<rfloor>" |
"\<langle>Suc n\<rangle> \<lparr>_\<rparr> = \<lfloor>\<rfloor>" |
"\<langle>Suc n\<rangle> ( _ \<Colon> \<tau>) = \<langle>n\<rangle> \<tau>"


lemma nth_suffix_nth: "\<langle>n\<rangle>\<tau> = \<lfloor>\<tau>'\<rfloor> \<Longrightarrow> \<tau>'\<langle>m\<rangle> = \<lfloor>s\<rfloor> \<Longrightarrow>  \<tau>\<langle>n+m\<rangle> = \<lfloor>s\<rfloor>"   
proof (induction n arbitrary: \<tau> \<tau>' m s)
case 0
  then show ?case
    by (simp add: some_def) 
next
  case (Suc n)
  then show ?case
  proof (cases \<tau>)
    case (one s')
    then show ?thesis
      by (metis Suc.prems(1) none_def option.distinct(1) some_def suffix.simps(2))
  next
  case (add s' \<tau>')
  then show ?thesis
    using Suc.IH Suc.prems(1) Suc.prems(2) by auto 
  qed
qed


lemma nth_Some_suffix :  "\<tau>\<langle>n\<rangle> = \<lfloor>s'\<rfloor> \<Longrightarrow> \<exists> \<tau>'. \<langle>n\<rangle> \<tau> = \<lfloor>\<tau>'\<rfloor>"
proof (induction n arbitrary: \<tau> s')
case 0
  then show ?case by auto
next
  case (Suc n)
  then show ?case
    by (metis Zero_not_Suc add_diff_cancel_left' none_def nth.elims option.distinct(1)
        plus_1_eq_Suc some_def suffix.simps(3))
qed

lemma head_nth_suffix : "\<langle>n\<rangle> \<tau> = \<lfloor>\<tau>'\<rfloor> \<Longrightarrow> \<tau>\<langle>n\<rangle> = \<lfloor>s\<rfloor> \<Longrightarrow>   s = \<alpha> \<tau>'"
proof (induction n arbitrary: \<tau>)
case 0
  then show ?case
    by (simp add: some_def) 
next
  case (Suc n)
  then show ?case
    by (metis seq.inject(2) Zero_not_Suc add_diff_cancel_left' none_def nth.elims
        option.distinct(1) plus_1_eq_Suc some_def suffix.elims)
  
qed
lemma isFinite_suffix : "\<propto> \<tau> \<Longrightarrow> \<langle>n\<rangle>\<tau> = \<lfloor>\<tau>'\<rfloor> \<Longrightarrow>\<propto> \<tau>'"
proof (induction n arbitrary: \<tau> \<tau>')
  case 0
  then show ?case
    by (metis option.simps(1) some_def suffix.simps(1)) 
next
  case (Suc n)
  then show ?case
    by (metis isFinite.cases none_def option.distinct(1) some_def suffix.simps(2) suffix.simps(3))
qed  
 
coinductive isPath ("\<pi>") where 
path_one : "\<bullet> s \<Longrightarrow> \<pi> \<lparr>s\<rparr>" |
path_add : "s \<mapsto> \<alpha> \<tau> \<Longrightarrow> \<pi> \<tau> \<Longrightarrow> \<pi> (s \<Colon> \<tau>)"

lemma isPath_suffix : "\<pi> \<tau> \<Longrightarrow> \<langle>n\<rangle>\<tau> = \<lfloor>\<tau>'\<rfloor> \<Longrightarrow> \<pi> \<tau>'"
proof (induction n arbitrary: \<tau> \<tau>')
case 0
  then show ?case
  by (metis option.simps(1) some_def suffix.simps(1)) 
next
  case (Suc n)
  then show ?case
    by (metis seq.distinct(1) Zero_not_Suc isPath.simps none_def option.distinct(1) some_def 
        suffix.elims suffix.simps(3))  
qed


coinductive reach :: "seq \<Rightarrow> SymState \<Rightarrow> bool" (infix "\<leadsto>" 300)where
reach_now_one : "r s \<Longrightarrow>  \<lparr>s\<rparr> \<leadsto> r" |
reach_now_add : "r s \<Longrightarrow> (s \<Colon> _) \<leadsto> r" |
reach_later : "\<tau> \<leadsto> r \<Longrightarrow>  ( _\<Colon> \<tau>) \<leadsto> r"

coinductive reach' :: "seq \<Rightarrow> SymState \<Rightarrow> bool" (infix "\<hookrightarrow>" 300)where
reach_now_one' : "r s \<Longrightarrow>  \<lparr>s\<rparr> \<hookrightarrow> r" |
reach_now_add' : "r s \<Longrightarrow> (s \<Colon> _) \<hookrightarrow> r" |
reach_later' : "\<langle>n\<rangle>\<tau> = \<lfloor>\<tau>'\<rfloor> \<Longrightarrow>  \<tau>' \<hookrightarrow> r \<Longrightarrow>  ( _\<Colon> \<tau>) \<hookrightarrow> r"

lemma reach_reach' : "\<tau> \<leadsto> r \<Longrightarrow>  \<tau> \<hookrightarrow> r"
proof (coinduction arbitrary: \<tau>)
  case reach'
  then show ?case 
  proof (cases \<tau>)
    case (reach_now_one s)
then show ?thesis
  by auto 
next
  case (reach_now_add s \<tau>')
  then show ?thesis
    by simp 
next
  case (reach_later \<tau>' s)
  then show ?thesis
    using suffix.simps(1) by fastforce
    print_theorems
  qed
qed

lemma reach'_reach : "\<tau> \<hookrightarrow> r \<Longrightarrow> \<tau> \<leadsto> r "
proof (coinduction arbitrary: \<tau>)
  case reach
  then show ?case 
  proof (cases \<tau>)
    case (reach_now_one' s)
then show ?thesis
  by simp 
next
  case (reach_now_add' s \<tau>'')
  then show ?thesis
    by simp 
next
  case (reach_later' n \<tau>'' \<tau>''' s)
  then show ?thesis
    by (metis none_def option.distinct(1) option.sel reach'.reach_later' some_def suffix.elims) 
qed
qed

lemma isFinite_reach_ex : "\<propto> \<tau> \<Longrightarrow> \<tau> \<leadsto> r \<Longrightarrow> \<exists> n s.  \<tau>\<langle>n\<rangle> = \<lfloor>s\<rfloor> \<and> r s"
proof (induction rule:isFinite.induct )
  case (finite_one s)
  then show ?case
    by (metis nth.simps(1) seq.distinct(1) reach.cases seq.sel(1)) 
next
  case (finite_add  \<tau> s)
  then show ?case
    by (metis nth.simps(1) nth.simps(3) seq.distinct(1) seq.inject(2) 
        seq.simps(6) \<alpha>_def reach.simps) 
qed
 

lemma nth_reach :   "\<tau>\<langle>n\<rangle> = \<lfloor>s\<rfloor> \<Longrightarrow> r s \<Longrightarrow> \<tau> \<leadsto> r"
proof (induction n arbitrary:\<tau>  s)
case 0
  then show ?case
    by (metis nth.simps(1) head_nth_suffix nth_Some_suffix reach.simps seq.exhaust_sel)
  
next
  case (Suc n)
  then show ?case
    by (metis Zero_not_Suc add_diff_cancel_left' none_def nth.elims option.distinct(1) 
        plus_1_eq_Suc reach.simps some_def)
qed

datatype reachability_formula =
 arrow (L:SymState) (R:SymState) (infix "\<Rightarrow>\<diamondsuit>" 150)

definition valid :: "reachability_formula \<Rightarrow> bool" ("\<Turnstile>") where
"\<Turnstile> f \<equiv> \<forall> \<tau>. (L f) (\<alpha> \<tau>) \<longrightarrow> \<pi> \<tau> \<longrightarrow>\<propto> \<tau> \<longrightarrow> \<tau> \<leadsto>(R f)"


lemma valid_finite_ex : 
  "\<Turnstile> (l \<Rightarrow>\<diamondsuit> r) \<Longrightarrow>
   \<forall>\<tau>.  (l (\<alpha> \<tau>) \<longrightarrow> \<pi> \<tau> \<longrightarrow> \<propto> \<tau> \<longrightarrow> (\<exists> n s.  \<tau>\<langle>n\<rangle> = \<lfloor>s\<rfloor> \<and> r s))"
  by (simp add: isFinite_reach_ex valid_def)

lemma ex_finite_valid :
 " (\<forall>\<tau>.  l(\<alpha> \<tau>) \<longrightarrow> \<pi> \<tau> \<longrightarrow> \<propto> \<tau> \<longrightarrow> (\<exists> n s.   \<tau>\<langle>n\<rangle> = \<lfloor>s\<rfloor> \<and> r s)) 
\<Longrightarrow> \<Turnstile> (l \<Rightarrow>\<diamondsuit> r)"
  using nth_reach valid_def by auto

inductive prove_ind :: "(reachability_formula \<Rightarrow> bool) \<Rightarrow>
                        (reachability_formula \<Rightarrow> bool) \<Rightarrow> 
                         reachability_formula \<Rightarrow> bool" ("_\<turnstile>__"  [910,0] 950) for H \<rho>   
  where
prove_ind_hyp : "H \<phi> \<Longrightarrow> H  \<turnstile> \<rho> \<phi>  " |
prove_ind_trv :  "H \<turnstile>  \<rho> (r \<Rightarrow>\<diamondsuit> r)" | 
prove_ind_str : "l \<sqsubseteq> l' \<Longrightarrow> 
                H \<turnstile>  \<rho> (l' \<Rightarrow>\<diamondsuit> r) \<Longrightarrow> 
                H \<turnstile>  \<rho> (l \<Rightarrow>\<diamondsuit> r)" |
prove_ind_dsj : "H \<turnstile> \<rho> (l1 \<Rightarrow>\<diamondsuit> r) \<Longrightarrow> 
                 H \<turnstile> \<rho> (l2 \<Rightarrow>\<diamondsuit> r) \<Longrightarrow> 
                 H \<turnstile> \<rho> ((l1 \<squnion> l2) \<Rightarrow>\<diamondsuit> r)" |
prove_ind_tra : "\<Turnstile>(l \<Rightarrow>\<diamondsuit> m) \<Longrightarrow> 
                H  \<turnstile> \<rho> (m \<Rightarrow>\<diamondsuit> r) \<Longrightarrow> 
                H \<turnstile> \<rho> (l \<Rightarrow>\<diamondsuit> r)" |
prove_ind_stp : "(l \<sqinter> \<bullet>) \<sqsubseteq> \<bottom> \<Longrightarrow>
                 \<rho> (\<partial> l \<Rightarrow>\<diamondsuit> r) \<Longrightarrow>
                H \<turnstile> \<rho> (l \<Rightarrow>\<diamondsuit> r)"


definition prove :: "(reachability_formula \<Rightarrow>bool) \<Rightarrow> reachability_formula \<Rightarrow>bool" ("_\<tturnstile>_" [911,0] 951)
  where "H \<tturnstile> \<phi> \<equiv> (gfp (prove_ind H)) \<phi>" 


lemma mono_proof_ind_aux : assumes " r1 \<le> r2"
  shows "H \<turnstile> r1 (l  \<Rightarrow>\<diamondsuit>r) \<Longrightarrow> H \<turnstile> r2 (l \<Rightarrow>\<diamondsuit> r)"
proof (induction rule:prove_ind.induct)
case (prove_ind_hyp r)
  then show ?case
    by (simp add: prove_ind.prove_ind_hyp) 
next
case (prove_ind_trv r)
  then show ?case
    by (simp add: prove_ind.prove_ind_trv) 
next
case (prove_ind_str l l' r)
  then show ?case
    using prove_ind.prove_ind_str by blast 
next
case (prove_ind_dsj l1 r l2)
  then show ?case
    by (simp add: prove_ind.prove_ind_dsj) 
next
case (prove_ind_tra l m r)
  then show ?case
    using prove_ind.prove_ind_tra by blast
next
case (prove_ind_stp l r)
  then show ?case
    using assms prove_ind.prove_ind_stp by auto 
qed

lemma mono_proof_ind : "mono (prove_ind H)"
proof (rule ;rule)
  show "\<And>x y xa. x \<le> y \<Longrightarrow> H \<turnstile> x xa \<Longrightarrow>H \<turnstile> y xa "
    by (metis (full_types) mono_proof_ind_aux reachability_formula.exhaust_sel)
qed

lemma icofix :
  fixes H f
  assumes Hpf : "P f"
  assumes Hpg : "(\<And>g. P g \<Longrightarrow> H  \<turnstile> (\<lambda>h. P h ) g)" 
  shows "H \<tturnstile> f"
  by (smt Hpf Hpg coinduct dual_order.refl le_boolI' le_funI le_sup_iff mono_def
      mono_proof_ind prove_def rev_predicate1D)




lemma prove_gfp :  " H \<tturnstile> f \<Longrightarrow>  H \<turnstile> (gfp (prove_ind H)) f"
  by (metis gfp_unfold mono_proof_ind prove_def)

lemma prove_intros : "H \<turnstile> (gfp (prove_ind H))  f \<Longrightarrow> H \<tturnstile> f"
  by (metis gfp_unfold mono_proof_ind prove_def)

lemma prove_cases : "H \<tturnstile> f \<Longrightarrow> (\<And>g. f = g \<Longrightarrow>H \<turnstile> (gfp (prove_ind H)) g \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: prove_gfp)

lemma my_reach'_coinduct : 
"X \<tau> r \<Longrightarrow>
(\<And> \<tau>' r'.
    X  \<tau>' r' \<Longrightarrow>
    (\<exists> s.  \<tau>' = \<lparr>s\<rparr> \<and> r' s) \<or>
    (\<exists> s \<tau>''.  \<tau>' = (s\<Colon>\<tau>'')  \<and> r' s) \<or> 
    (\<exists>n \<tau>'' \<tau>''' s.  \<tau>' = (s\<Colon>\<tau>'') \<and> \<langle>n\<rangle>\<tau>'' = \<lfloor>\<tau>'''\<rfloor> \<and> (X \<tau>''' r' \<or> \<tau>''' \<hookrightarrow> r'))) \<Longrightarrow>
\<tau> \<hookrightarrow> r"
  by (smt reach'.coinduct)


lemma soundness_aux2  : 
  assumes valid_hyps : "\<forall>f. (H f \<longrightarrow> \<Turnstile> f)"
  shows
"H \<turnstile> (gfp (prove_ind H)) f  \<Longrightarrow> (L f) (\<alpha> \<tau>) \<Longrightarrow> \<pi> \<tau> \<Longrightarrow> \<propto> \<tau> \<Longrightarrow>
       (\<exists>s. \<tau> = \<lparr>s\<rparr> \<and> (R f) s) \<or>
       (\<exists>s \<tau>'. \<tau> = (s\<Colon> \<tau>') \<and> (R f) s) \<or>
       (\<exists>n s \<tau>' \<tau>''.
           \<tau> = (s\<Colon> \<tau>') \<and> \<langle>n\<rangle>\<tau>' = \<lfloor>\<tau>''\<rfloor> \<and>
           (\<exists>l'.( (H \<tturnstile>(l'\<Rightarrow>\<diamondsuit>(R f))) \<and> l' (\<alpha> \<tau>'') \<and>  \<pi> \<tau>'' \<and> \<propto> \<tau>'') \<or>  \<tau>''\<hookrightarrow> (R f)))"
proof (induction arbitrary: \<tau> rule:prove_ind.induct)
case (prove_ind_hyp f)
  then show ?case 
  proof -
    have " \<Turnstile> f"
      by (simp add: prove_ind_hyp.hyps valid_hyps) 
    hence
     " (L f) (\<alpha> \<tau>) \<Longrightarrow> \<pi> \<tau> \<Longrightarrow> \<propto> \<tau> \<Longrightarrow>
       (\<exists>s. \<tau> = \<lparr>s\<rparr> \<and> (R f) s) \<or>
       (\<exists>s \<tau>'. \<tau> = (s\<Colon> \<tau>') \<and> (R f) s) \<or>
       (\<exists>n s \<tau>' \<tau>''.
           \<tau> = (s\<Colon> \<tau>') \<and> \<langle>n\<rangle>\<tau>' = \<lfloor>\<tau>''\<rfloor> \<and>   \<tau>''\<hookrightarrow> (R f))"
      using reach'.cases reach_reach' valid_def
      by (metis Proof_System_2.seq.distinct(1)  prove_ind_hyp.prems(1) prove_ind_hyp.prems(2) 
          prove_ind_hyp.prems(3) reach.simps reachability_formula.exhaust reachability_formula.sel(2) 
          seq.sel(1)  suffix.simps(1))
    thus ?thesis
      using prove_ind_hyp.prems(1) prove_ind_hyp.prems(2) prove_ind_hyp.prems(3) by blast 
  qed
next
case (prove_ind_trv r)
  then show ?case
    by (metis reachability_formula.sel(1) reachability_formula.sel(2) seq.exhaust_sel) 
next
case (prove_ind_str l l' r)
  then show ?case 
    using imp_def by fastforce
next
  case (prove_ind_dsj l1 r l2)
  then show ?case 
    using dsj_def by fastforce
next
  case (prove_ind_tra l m r)
  then show ?case 
  proof -
    obtain n s where Htaun : "\<tau>\<langle>n\<rangle> = \<lfloor>s\<rfloor>" and Hms : "m s"
      using prove_ind_tra.hyps(1) prove_ind_tra.prems(1) 
        prove_ind_tra.prems(2) prove_ind_tra.prems(3) valid_finite_ex by fastforce
    then obtain \<tau>' where Hsuf: "\<langle>n\<rangle>\<tau> = \<lfloor>\<tau>'\<rfloor>" and Hinit: "m (\<alpha> \<tau>')" and Hpath: "\<pi> \<tau>'" and Hfin: "\<propto> \<tau>'"
      using head_nth_suffix isFinite_suffix isPath_suffix nth_Some_suffix prove_ind_tra.prems(2)
            prove_ind_tra.prems(3) by fastforce
    hence Hbig : "(\<exists>s. \<tau>' = \<lparr>s\<rparr> \<and> r s) \<or>
      (\<exists>s \<tau>''. \<tau>' = (s\<Colon>\<tau>'') \<and> r s) \<or>
      (\<exists>n s \<tau>'' \<tau>'''. \<tau>' =(s\<Colon>\<tau>'') \<and> \<langle>n\<rangle>\<tau>'' = \<lfloor>\<tau>'''\<rfloor> \<and> 
       (\<exists>l'. (H \<tturnstile> (l' \<Rightarrow>\<diamondsuit> r)) \<and> l' (\<alpha> \<tau>''') \<and> \<pi> \<tau>''' \<and> \<propto> \<tau>''' \<or> \<tau>''' \<hookrightarrow> r))"
      using prove_ind_tra.IH  by auto
    thus ?thesis
    proof (cases \<tau>')
      case (one s')
      then show ?thesis
        by (metis Hsuf seq.distinct(1) Hbig none_def option.distinct(1) option.sel reach'.intros(1)
            reachability_formula.sel(2) some_def suffix.elims)
    next
      case (add s' \<tau>'')
      then show ?thesis
        by (metis Hbig Hfin Hinit Hpath Hsuf none_def option.discI option.inject prove_intros
            prove_ind_tra.hyps(2) reachability_formula.sel(2) some_def suffix.elims) 
    qed
  qed
next
  case (prove_ind_stp l r)
  then show ?case 
  proof (cases \<tau>)
    case (one s)
    then show ?thesis 
      by (metis seq.distinct(1) bot_def cnj_def imp_def isPath.simps prove_ind_stp.hyps(1) 
          prove_ind_stp.prems(1) prove_ind_stp.prems(2) reachability_formula.sel(1) seq.sel(1)) 
  next
     case (add s \<tau>')
     then show ?thesis 
     proof (cases "r s")
       case True
       then show ?thesis
         by (simp add: add) 
     next
       case False
       then show ?thesis  
       proof -
         have "(\<exists> s \<tau>''.
        \<tau> = (s\<Colon>\<tau>'') \<and>  ((H\<tturnstile>(\<partial>l \<Rightarrow>\<diamondsuit> r)) \<and> (\<partial> l) (\<alpha> \<tau>'') \<and> \<pi> \<tau>'' \<and> \<propto> \<tau>''))"
           by (metis Proof_System_2.seq.simps(4) add isFinite.cases isPath.simps prove_def prove_ind_stp.hyps(2) prove_ind_stp.prems(1) prove_ind_stp.prems(2) prove_ind_stp.prems(3) 
             reachability_formula.sel(1) seq.sel(2) seq.sel(3) symTrans_def)
          hence "(\<exists> n s \<tau>'' \<tau>'''.
        \<tau> = (s\<Colon>\<tau>'') \<and>  \<langle>n\<rangle>\<tau>'' = \<lfloor>\<tau>'''\<rfloor> \<and>   ((H\<tturnstile>(\<partial>l \<Rightarrow>\<diamondsuit> r)) \<and> (\<partial> l) (\<alpha> \<tau>''') \<and> \<pi> \<tau>''' \<and> \<propto> \<tau>'''))"
            using suffix.simps(1) by blast
         thus ?thesis  
           by auto
     qed
   qed
 qed
qed

lemma soundness_aux1 :fixes H
  assumes valid_hyps : "\<forall>f. (H f \<longrightarrow> \<Turnstile> f)"
  shows  " (H\<tturnstile>(l' \<Rightarrow>\<diamondsuit> r')) \<and> l' (\<alpha> \<tau>') \<and> \<pi> \<tau>' \<and> \<propto> \<tau>' \<Longrightarrow>
  (\<exists>s. \<tau>' = \<lparr>s\<rparr> \<and> r' s) \<or>
  (\<exists>s \<tau>''. \<tau>' = (s\<Colon>\<tau>'') \<and> r' s) \<or>     
  (\<exists>n \<tau>'' \<tau>''' s. \<tau>' = (s\<Colon>\<tau>'') \<and> \<langle>n\<rangle>\<tau>'' = \<lfloor>\<tau>'''\<rfloor> \<and> ((\<exists>l''. (H\<tturnstile>(l'' \<Rightarrow>\<diamondsuit> r')) \<and> l'' (\<alpha> \<tau>''') \<and> \<pi> \<tau>''' \<and> \<propto> \<tau>''') \<or> \<tau>''' \<hookrightarrow> r'))"
proof -
  assume "(H\<tturnstile>(l' \<Rightarrow>\<diamondsuit> r')) \<and> l' (\<alpha> \<tau>') \<and> \<pi> \<tau>' \<and> \<propto> \<tau>'"
  hence Hhyp: "(H \<turnstile> (gfp (prove_ind H)) (l' \<Rightarrow>\<diamondsuit> r') ) \<and> l' (\<alpha> \<tau>') \<and> \<pi> \<tau>' \<and> \<propto> \<tau>' "
    using prove_gfp by auto
  thus ?thesis  using soundness_aux2 [of H "(l' \<Rightarrow>\<diamondsuit> r')" \<tau>']
    using valid_hyps by auto 
  hence " (\<exists>s. \<tau>' = \<lparr>s\<rparr> \<and> r' s) \<or>
 (\<exists>s \<tau>''. \<tau>' = (s\<Colon>\<tau>'') \<and> r' s) \<or>     
 (\<exists>n \<tau>'' \<tau>''' s. \<tau>' = (s\<Colon>\<tau>'') \<and> \<langle>n\<rangle>\<tau>'' = \<lfloor>\<tau>'''\<rfloor> \<and> ((\<exists>l''. (H\<tturnstile>(l'' \<Rightarrow>\<diamondsuit> r')) \<and> l'' (\<alpha> \<tau>''') \<and> \<pi> \<tau>''' \<and> \<propto> \<tau>''') \<or> \<tau>''' \<hookrightarrow> r'))"
    using soundness_aux2 [of H "(l' \<Rightarrow>\<diamondsuit> r')" \<tau>'] valid_hyps by auto
qed

lemma soundness_aux0 :fixes H
  assumes valid_hyps : "\<forall>f. (H f \<longrightarrow> \<Turnstile> f)"
  shows  " \<exists>l'.(H\<tturnstile>(l' \<Rightarrow>\<diamondsuit> r')) \<and> l' (\<alpha> \<tau>') \<and> \<pi> \<tau>' \<and> \<propto> \<tau>' \<Longrightarrow>
  (\<exists>s. \<tau>' = \<lparr>s\<rparr> \<and> r' s) \<or>
  (\<exists>s \<tau>''. \<tau>' = (s\<Colon>\<tau>'') \<and> r' s) \<or>     
  (\<exists>n \<tau>'' \<tau>''' s. \<tau>' = (s\<Colon>\<tau>'') \<and> \<langle>n\<rangle>\<tau>'' = \<lfloor>\<tau>'''\<rfloor> \<and> ((\<exists>l''. (H\<tturnstile>(l'' \<Rightarrow>\<diamondsuit> r')) \<and> l'' (\<alpha> \<tau>''') \<and> \<pi> \<tau>''' \<and> \<propto> \<tau>''') \<or> \<tau>''' \<hookrightarrow> r'))"
  using soundness_aux1
  using valid_hyps by blast


lemma soundness_aux : 
  fixes H
  assumes valid_hyps : "\<forall>f. (H f \<longrightarrow> \<Turnstile> f)"
  shows 
 "H \<tturnstile> (l\<Rightarrow>\<diamondsuit>r) \<Longrightarrow> l (\<alpha> \<tau>) \<Longrightarrow> \<pi> \<tau> \<Longrightarrow> \<propto> \<tau> \<Longrightarrow> \<tau> \<hookrightarrow>r"
proof(intro my_reach'_coinduct [of "\<lambda>\<tau>' r'.\<exists>l'. (H \<tturnstile> (l'\<Rightarrow>\<diamondsuit>r')) \<and> l' (\<alpha> \<tau>') \<and> \<pi> \<tau>' \<and> \<propto> \<tau>'"] )
  show
 " H \<tturnstile>(l \<Rightarrow>\<diamondsuit> r) \<Longrightarrow> l (\<alpha> \<tau>) \<Longrightarrow> \<pi> \<tau> \<Longrightarrow> \<propto> \<tau> \<Longrightarrow> \<exists>l. (H \<tturnstile> (l \<Rightarrow>\<diamondsuit> r)) \<and> l (\<alpha> \<tau>) \<and> \<pi> \<tau> \<and> \<propto> \<tau>" by blast
  then show 
   "\<And>\<tau>' r'.
       H\<tturnstile>(l \<Rightarrow>\<diamondsuit> r) \<Longrightarrow>
       l (\<alpha> \<tau>) \<Longrightarrow>
       \<pi> \<tau> \<Longrightarrow>
       \<propto> \<tau> \<Longrightarrow>
       \<exists>l'. (H\<tturnstile>(l' \<Rightarrow>\<diamondsuit> r')) \<and> l' (\<alpha> \<tau>') \<and> \<pi> \<tau>' \<and> \<propto> \<tau>' \<Longrightarrow>
       (\<exists>s. \<tau>' = \<lparr>s\<rparr> \<and> r' s) \<or>
       (\<exists>s \<tau>''. \<tau>' = (s\<Colon>\<tau>'') \<and> r' s) \<or>
       (\<exists>n \<tau>'' \<tau>''' s. \<tau>' = (s\<Colon>\<tau>'') \<and> \<langle>n\<rangle>\<tau>'' = \<lfloor>\<tau>'''\<rfloor> \<and> ((\<exists>l'. (H\<tturnstile>(l' \<Rightarrow>\<diamondsuit> r')) \<and> l' (\<alpha> \<tau>''') \<and> \<pi> \<tau>''' \<and> \<propto> \<tau>''') \<or> \<tau>''' \<hookrightarrow> r'))"
    using soundness_aux0
    by (simp add: valid_hyps)
qed


theorem soundness : 
assumes  "\<forall>f. (H f \<longrightarrow> \<Turnstile> f)"
shows "H \<tturnstile> f \<Longrightarrow> \<Turnstile> f"
  by (metis assms reach'_reach reachability_formula.collapse soundness_aux valid_def)


lemma prove_intros_bis : "H \<turnstile> ( (\<lambda>\<phi>. H \<tturnstile> \<phi>))  f \<Longrightarrow> H \<tturnstile> f"
  by (metis (full_types) dual_order.antisym  gfp_unfold mono_proof_ind predicate1I prove_def)


lemma strategy_inv : 
  assumes Hinit : "l\<sqsubseteq> q \<squnion> r"
  and Hnfinal : "q \<sqinter> \<bullet> \<sqsubseteq> \<bottom>"
  and Hqinv : "\<partial>q \<sqsubseteq> q \<squnion>r"
  shows "(\<lambda>f. False) \<tturnstile> (l \<Rightarrow>\<diamondsuit> r)"
  apply (rule prove_intros_bis)
  apply (rule prove_ind_str [of _ "q \<squnion> r"])
      apply (rule Hinit)
      apply (rule prove_ind_dsj)
          apply (rule prove_ind_stp)  
          apply (rule Hnfinal) 
              apply (rule icofix [of  "\<lambda> f. f = (\<partial>q \<Rightarrow>\<diamondsuit> r)"]; auto) 
              apply (rule prove_ind_str [of _ "q \<squnion> r"])
                  apply (rule Hqinv)
                  apply (rule prove_ind_dsj)
                      apply (rule prove_ind_stp)  
                          apply (rule Hnfinal)
                          apply (auto)
                  apply (rule prove_ind_trv)
          apply (rule prove_ind_trv)
  done

definition coReach :: "SymState \<Rightarrow> SymState" ("\<circle>\<lozenge>") where
"\<circle>\<lozenge>r \<equiv> \<lambda> s.\<not>r s \<and> (\<forall> \<tau>. s = (\<alpha> \<tau>) \<longrightarrow> \<pi> \<tau> \<longrightarrow> \<propto> \<tau> \<longrightarrow> \<tau> \<leadsto>r)"

lemma init : "\<Turnstile> (l\<Rightarrow>\<diamondsuit>r) \<Longrightarrow>  l \<sqsubseteq> \<circle>\<lozenge>r \<squnion> r"
  using coReach_def dsj_def imp_def reach_reach' valid_def by auto

lemma nonBlock : "\<circle>\<lozenge>r \<sqinter> \<bullet> \<sqsubseteq> \<bottom>"
proof (cases "\<exists>s. \<circle>\<lozenge> r s \<and> \<bullet> s")
  case True
  thus ?thesis 
  proof -
    have False
      by (metis seq.distinct(1) True coReach_def finite_one path_one reach.cases seq.sel(1))
    thus ?thesis by simp
  qed
next
  case False
  thus ?thesis
    using cnj_def imp_def by auto 
qed

lemma quasiInv : "\<partial>(\<circle>\<lozenge>r) \<sqsubseteq> \<circle>\<lozenge>r \<squnion> r"
proof (cases "\<exists> s.  \<partial>(\<circle>\<lozenge>r) s \<and> \<not> \<circle>\<lozenge>r s \<and> \<not> r s")
  case True
  then show ?thesis 
  proof -
    obtain s where " \<partial>(\<circle>\<lozenge>r) s" and  "\<not>(\<circle>\<lozenge>r) s"  and  "\<not> r s" using True 
      by blast
    then obtain \<tau> where "s = (\<alpha> \<tau>)" and "\<pi> \<tau>" and "\<propto> \<tau>" and "\<not> \<tau> \<leadsto>r"
      using coReach_def by auto
    obtain s' where "(\<circle>\<lozenge> r) s'" and "s'\<mapsto> s"
      using \<open> \<partial>(\<circle>\<lozenge>r) s\<close> symTrans_def by auto
    have "\<not> r s'" and "\<pi> (s'\<Colon> \<tau>)" and "\<propto> (s'\<Colon> \<tau>)"
      proof -
        show "\<not> r s'"
        using \<open>\<circle>\<lozenge>r s'\<close> coReach_def by force
      next show "\<pi> (s'\<Colon> \<tau>)"
          using \<open>s' \<mapsto> s\<close> \<open>\<pi> \<tau>\<close> \<open>s = \<alpha> \<tau>\<close> path_add by blast
      next show "\<propto> (s'\<Colon> \<tau>)"
          by (simp add: \<open>\<propto> \<tau>\<close> finite_add)
    qed
    have " (s'\<Colon> \<tau>) \<leadsto> r"
      using \<open>\<pi> (s'\<Colon>\<tau>)\<close> \<open>\<propto> (s'\<Colon>\<tau>)\<close> \<open>\<circle>\<lozenge>r s'\<close> coReach_def by auto
    have False
      using \<open>\<not> r s'\<close> \<open>\<not> \<tau> \<leadsto> r\<close> \<open>(s'\<Colon>\<tau>) \<leadsto> r\<close> reach.simps by blast
    thus ?thesis by auto
  qed
next
case False
  then show ?thesis
    using dsj_def imp_def by auto 
qed

theorem completeness : "\<Turnstile> f \<Longrightarrow>(\<lambda>\<phi>. False) \<tturnstile> f "
  by (metis (no_types) prove_intros prove_ind_tra prove_ind_trv reachability_formula.exhaust)

theorem compositionality : "(\<lambda>\<phi>. False) \<tturnstile> f'  \<Longrightarrow> (\<lambda>\<phi>. \<phi> = f') \<tturnstile> f \<Longrightarrow> (\<lambda>\<phi>. False) \<tturnstile> f"
  by (metis (full_types) completeness soundness)
  
end
